import json
import math

from typing import List, Optional, Tuple
from urllib.parse import urlparse, parse_qs

from .exceptions import LnurlResponseException, InvalidLnurlTag, InvalidLnurlPayMetadata
from .utils import decode, snake_keys


class ParsedUrl:
    def __init__(self, url: str) -> None:
        self._parsed_url = urlparse(url)
        self.full = url
        self.base = f'{self._parsed_url.scheme}://{self._parsed_url.netloc}{self._parsed_url.path}'
        self.query_params = {k: v[0] for k, v in parse_qs(self._parsed_url.query).items()}

    def __repr__(self) -> str:
        return f'<ParsedUrl [{self.full}]>'

    def __str__(self) -> str:
        return self.full


class Lnurl:
    def __init__(self, lnurl: str, *, force_https: bool = True) -> None:
        self.bech32 = lnurl
        self.url = ParsedUrl(decode(lnurl, force_https=force_https))

    def __repr__(self) -> str:
        return f'<Lnurl [{self.bech32}]>'

    def __str__(self) -> str:
        return self.bech32

    @property
    def decoded(self) -> str:
        return self.url.full


class LnurlResponse:
    _tag = None

    def __init__(self, d: dict) -> None:
        self.__dict__ = snake_keys(d)

        if 'callback' in self.__dict__:
            self.callback = ParsedUrl(self.callback)

        if 'tag' in self.__dict__ and self.tag != self._tag:
            raise InvalidLnurlTag

    @property
    def ok(self) -> bool:
        """Returns True if :attr:`status` is "OK" or if we receive a tagged response from the server.
        """
        try:
            self.raise_for_status()
        except LnurlResponseException:
            return False
        return True

    @property
    def error_msg(self) -> Optional[str]:
        if 'status' in self.__dict__ and self.status == 'ERROR':
            return self.reason
        return None

    def raise_for_status(self) -> None:
        if self.error_msg:
            raise LnurlResponseException(self.error_msg)


class LnurlAuthResponse(LnurlResponse):
    _tag = 'login'

    def __init__(self, d) -> None:
        raise NotImplementedError


class LnurlChannelResponse(LnurlResponse):
    _tag = 'channelRequest'


class LnurlHostedChannelResponse(LnurlResponse):
    _tag = 'hostedChannelRequest'


class LnurlPayResponse(LnurlResponse):
    _tag = 'payRequest'
    _valid_metadata_mime_types = ['text/plain']

    def __init__(self, d: dict) -> None:
        super().__init__(d)

        if 'metadata' in self.__dict__:
            self.metadata = self.read_metadata(self.metadata)

    def read_metadata(self, json_str: str) -> List[Tuple]:
        try:
            return [tuple(x) for x in json.loads(json_str) if x[0] in self._valid_metadata_mime_types]
        except Exception:
            raise InvalidLnurlPayMetadata

    @property
    def min_sats(self) -> int:
        return int(math.ceil(self.min_sendable / 1000))

    @property
    def max_sats(self) -> int:
        return int(math.floor(self.max_sendable / 1000))


class LnurlWithdrawResponse(LnurlResponse):
    _tag = 'withdrawRequest'

    @property
    def min_sats(self) -> int:
        return int(math.ceil(self.min_withdrawable / 1000))

    @property
    def max_sats(self) -> int:
        return int(math.floor(self.max_withdrawable / 1000))



# flake8: noqa

from .models import (
    Lnurl,
    LnurlAuthResponse,
    LnurlChannelResponse,
    LnurlHostedChannelResponse,
    LnurlPayResponse,
    LnurlWithdrawResponse
)
from .utils import decode, encode



from os import path
from setuptools import setup


with open(path.join(path.abspath(path.dirname(__file__)), 'README.md')) as f:
    long_description = f.read()


setup(
    name='lnurl',
    version='0.1.0',
    url='https://github.com/python-ln/lnurl',
    author="jogco",
    author_email="jogco@lnsms.world",
    license='MIT',
    description='LNURL implementation for Python.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='bitcoin lightning-network lnurl',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Utilities',
    ],
    packages=['lnurl'],
    install_requires=[
        'bech32',
    ],
    zip_safe=False
)



import re

from bech32 import bech32_decode, bech32_encode, convertbits
from urllib.parse import urlparse

from .exceptions import InvalidLnurl, InvalidScheme, InvalidUrl


def validate_url(url: str, force_https: bool = True) -> None:
    try:
        parsed = urlparse(url)
    except ValueError:  # pragma: nocover
        raise InvalidUrl

    if force_https and parsed.scheme != 'https':
        raise InvalidScheme


def decode(lnurl: str, *, force_https: bool = True) -> str:
    lnurl = lnurl.replace('lightning:', '') if lnurl.startswith('lightning:') else lnurl
    hrp, data = bech32_decode(lnurl)

    if None in (hrp, data) or hrp != 'lnurl':
        raise InvalidLnurl

    try:
        url = bytes(convertbits(data, 5, 8, False)).decode('utf-8')
    except UnicodeDecodeError:  # pragma: nocover
        raise InvalidLnurl

    validate_url(url, force_https)
    return url


def encode(url: str, *, force_https: bool = True) -> str:
    validate_url(url, force_https)

    try:
        lnurl = bech32_encode('lnurl', convertbits(url.encode('utf-8'), 8, 5, True))
    except UnicodeEncodeError:  # pragma: nocover
        raise InvalidUrl

    return lnurl.upper()


def snake_keys(d: dict) -> dict:
    return {re.sub('([A-Z]+)', r'_\1', k).lower(): v for k, v in d.items()}
